# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-05-10 04:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0006_auto_20170509_0422'),
    ]

    operations = [
        migrations.AlterField(
            model_name='part',
            name='notes',
            field=models.TextField(blank=True, default='', help_text='Notes regarding this part.'),
        ),
    ]
